#include "AS3935.hpp"

// C++ system level
// #include <functional>

// ESP32 specific
#include "esp_log.h"
#include "esp_err.h"
#include "driver/i2c.h"
#include "driver/gpio.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

// project specific
#include <SpiderLib.hpp>

// qthing stuff
// #include <qthing>

// misc
// #include ""




static const char* TAG = "AS3935";


namespace AS3935 {


  AS3935::AS3935(u8 SDA, u8 SCL, u8 IRQ, u8 i2cAddr, u8 i2cChannel, u32 i2cFrequency) : ioIRQ(IRQ), i2cAddr(i2cAddr & I2C_ADDR_MASK), i2cChannel(i2cChannel) {

    i2c_config_t cfg;

    cfg.mode             = I2C_MODE_MASTER;
    cfg.sda_io_num       = (gpio_num_t)SDA;
    cfg.scl_io_num       = (gpio_num_t)SCL;
    cfg.sda_pullup_en    = GPIO_PULLUP_ENABLE;
    cfg.scl_pullup_en    = GPIO_PULLUP_ENABLE;
    cfg.master.clk_speed = i2cFrequency;

    i2c_param_config((i2c_port_t)i2cChannel, &cfg);

    esp_err_t err = i2c_driver_install((i2c_port_t)i2cChannel, cfg.mode, 0, 0, 0);

    ESP_LOGI(TAG, "I²C_driver_install(…): %s", err == ESP_OK ? "✅" : "❌");


    gpio_config_t gpio;

    gpio.mode         = GPIO_MODE_INPUT;
    gpio.intr_type    = GPIO_INTR_DISABLE;
    gpio.pull_up_en   = GPIO_PULLUP_DISABLE;
    gpio.pull_down_en = GPIO_PULLDOWN_DISABLE;
    gpio.pin_bit_mask = 1ULL << IRQ;

    gpio_config(&gpio);

    bool result = resetToDefaultState();
    ESP_LOGI(TAG, "ResetSensorState(…): %s", result ? "✅" : "❌");

    // TODO: what to do next…?

  }


  void AS3935::mainLoop() {
    while (true) {


      vTaskDelay(100 / portTICK_PERIOD_MS);
    }
  }


  bool AS3935::clearStatistics() {
    return mutex.runReturn<bool> ([&]() {
      R0x02 reg;
      bool result = true;

      // reg.ClearStatistics = 1;
      // result &= i2cWriteByte(0x02, reg.value);

      reg.ClearStatistics = 0;
      result &= i2cWriteByte(0x02, reg.value);

      reg.ClearStatistics = 1;
      result &= i2cWriteByte(0x02, reg.value);

      return result;
    });
  }


  bool AS3935::calibrateRCO() {
    // TODO: we have to do more here!!
    return i2cWriteByteMutex(REG_CALIB_RCO, DIRECT_COMMAND);
  }


  bool AS3935::resetToDefaultState() {
    return i2cWriteByteMutex(REG_PRESET_DEFAULT, DIRECT_COMMAND);
  }


  u32 AS3935::measureFLCO() {
    return mutex.runReturn<u32> ([&]() {
      R0x08 reg;
      i2cReadByte(0x08, &reg.value);  // TODO: log error on failure and abort…?

      reg.Display_LCO  = true;
      reg.Display_SRCO = false;
      reg.Display_TRCO = false;

      // TODO: disable interrupt if enabled if implemented

      i2cWriteByte(0x08, reg.value);  // TODO: log error on failure and abort…?

      // TODO: do the actual frequency measuremen!!
      u32 result = 1337;

      reg.Display_LCO = false;
      i2cWriteByte(0x08, reg.value);  // TODO: log error on failure and abort…?

      return result;
    });
  }


  u32 AS3935::measureFSRCO() {
    // TODO: implement!
    return 1337;
  }

  u32 AS3935::measureFTRCO() {
    // TODO: implement!
    return 1337;
  }




  bool AS3935::displayLCO() {
    R0x08 reg;
    reg.Display_LCO  = true;
    reg.Display_SRCO = false;
    reg.Display_TRCO = false;  // TODO: y won't TRCO display??
    //reg.TuningCapacitor = 0xF;
    return i2cWriteByte(0x08, reg.value);
  }



  bool AS3935::i2cWriteByte(u8 regAddr, u8 value) {
    i2c_cmd_handle_t cmd = i2c_cmd_link_create();

    i2c_master_start(cmd);

    i2c_master_write_byte(cmd, writeAddr(), true);
    i2c_master_write_byte(cmd, regAddr, true);
    i2c_master_write_byte(cmd, value, true);

    i2c_master_stop(cmd);

    esp_err_t err = i2c_master_cmd_begin((i2c_port_t)i2cChannel, cmd, 1000 / portTICK_PERIOD_MS);
    i2c_cmd_link_delete(cmd);

    return err == ESP_OK;
  }

  bool AS3935::i2cReadByte(u8 regAddr, u8* value) {
    i2c_cmd_handle_t cmd = i2c_cmd_link_create();

    i2c_master_start(cmd);

    i2c_master_write_byte(cmd, writeAddr(), true);
    i2c_master_write_byte(cmd, regAddr, true);

    i2c_master_start(cmd);  // repeated start condition  ==>  direction change

    i2c_master_write_byte(cmd, readAddr(), true);
    i2c_master_read_byte(cmd, value, I2C_MASTER_LAST_NACK);

    i2c_master_stop(cmd);

    esp_err_t err = i2c_master_cmd_begin((i2c_port_t)i2cChannel, cmd, 1000 / portTICK_PERIOD_MS);
    i2c_cmd_link_delete(cmd);

    return err == ESP_OK;
  }


  bool AS3935::i2cWriteByteMutex(u8 regAddr, u8 value) {
    return mutex.runReturn<bool> ([&]() {
      return i2cWriteByte(regAddr, value);
    });
  }

  bool AS3935::i2cReadByteMutex(u8 regAddr, u8* value) {
    return mutex.runReturn<bool> ([&]() {
      return i2cReadByte(regAddr, value);
    });
  }


  u8 AS3935::readAddr() const {
    return (i2cAddr << 1) | 1;
  }

  u8 AS3935::writeAddr() const {
    return i2cAddr << 1;
  }


  bool AS3935::getIRQState() {
    return gpio_get_level((gpio_num_t)ioIRQ);
  }

}
